@@ -6,23 +6,32 @@
 /*                    New code by Bill Reiss 1993                     */
 /**********************************************************************/
 #include "tintin.h"
+#include "kbtree.h"
 #include "protos/action.h"
 #include "protos/globals.h"
 #include "protos/hash.h"
 #include "protos/hooks.h"
 #include "protos/llist.h"
 #include "protos/print.h"
+#include "protos/math.h"
 #include "protos/net.h"
 #include "protos/parse.h"
 #include "protos/run.h"
 #include "protos/unicode.h"
 #include "protos/user.h"
 #include "protos/utils.h"
-#include "protos/variables.h"
+#include "protos/string.h"
+#include "protos/vars.h"
+#include "ttyrec.h"
 #include <pwd.h>
 #include <fcntl.h>
 
-static void prepare_for_write(const char *command, const char *left, const char *right, const char *pr, char *result);
+
+#ifndef O_BINARY
+# define O_BINARY 0
+#endif
+
+static void cfcom(FILE *f, const char *command, const char *left, const char *right, const char *pr);
 extern void char_command(const char *arg, struct session *ses);
 
 /*******************************/
@@ -152,17 +161,13 @@ void unlink_command(const char *arg, struct session *ses)
 {
     char file[BUFFER_SIZE], temp[BUFFER_SIZE], lstr[BUFFER_SIZE];
 
-    if (*arg)
-    {
-        arg = get_arg_in_braces(arg, temp, 1);
-        substitute_vars(temp, file);
-        substitute_myvars(file, temp, ses);
-        expand_filename(temp, file, lstr);
-        unlink(lstr);
-    }
-    else
-        tintin_eprintf(ses, "#ERROR: valid syntax is: #unlink <filename>");
+    if (!*arg)
+        return tintin_eprintf(ses, "#ERROR: valid syntax is: #unlink <filename>");
 
+    arg = get_arg_in_braces(arg, temp, 1);
+    substitute_vars(temp, temp, ses);
+    expand_filename(temp, file, lstr);
+    unlink(lstr);
 }
 
 /*************************/
@@ -173,25 +178,21 @@ void deathlog_command(const char *arg, struct session *ses)
     FILE *fh;
     char fname[BUFFER_SIZE], text[BUFFER_SIZE], temp[BUFFER_SIZE], lfname[BUFFER_SIZE];
 
-    if (*arg)
+    if (!*arg)
+        return tintin_eprintf(ses, "#ERROR: valid syntax is: #deathlog <file> <text>");
+
+    arg = get_arg_in_braces(arg, temp, 0);
+    arg = get_arg_in_braces(arg, text, 1);
+    substitute_vars(temp, temp, ses);
+    expand_filename(temp, fname, lfname);
+    substitute_vars(text, text, ses);
+    if ((fh = fopen(lfname, "a")))
     {
-        arg = get_arg_in_braces(arg, temp, 0);
-        arg = get_arg_in_braces(arg, text, 1);
-        substitute_vars(temp, fname);
-        substitute_myvars(fname, temp, ses);
-        expand_filename(temp, fname, lfname);
-        substitute_vars(text, temp);
-        substitute_myvars(temp, text, ses);
-        if ((fh = fopen(lfname, "a")))
-        {
-            cfprintf(fh, "%s\n", text);
-            fclose(fh);
-        }
-        else
-            tintin_eprintf(ses, "#ERROR: COULDN'T OPEN FILE {%s}.", fname);
+        cfprintf(fh, "%s\n", text);
+        fclose(fh);
     }
     else
-        tintin_eprintf(ses, "#ERROR: valid syntax is: #deathlog <file> <text>");
+        tintin_eprintf(ses, "#ERROR: COULDN'T OPEN FILE {%s}.", fname);
 }
 
 void log_off(struct session *ses)
@@ -285,15 +286,11 @@ void logcomment_command(const char *arg, struct session *ses)
     char text[BUFFER_SIZE];
 
     if (!arg)
-    {
-        tintin_eprintf(ses, "#Logcomment what?");
-        return;
-    }
+        return tintin_eprintf(ses, "#Logcomment what?");
+
     if (!ses->logfile)
-    {
-        tintin_eprintf(ses, "#You're not logging.");
-        return;
-    }
+        return tintin_eprintf(ses, "#You're not logging.");
+
     arg=get_arg(arg, text, 1, ses);
     write_logf(ses, text, "", "");
 }
@@ -407,30 +404,25 @@ static FILE* open_logfile(struct session *ses, const char *name, const char *fil
 void condump_command(const char *arg, struct session *ses)
 {
     FILE *fh;
-    char fname[BUFFER_SIZE], temp[BUFFER_SIZE];
+    char temp[BUFFER_SIZE];
 
     if (!ui_con_buffer)
+        return tintin_eprintf(ses, "#UI: No console buffer => can't dump it.");
+
+    if (!*arg)
+        return tintin_eprintf(ses, "#Syntax: #condump <file>");
+
+    arg = get_arg_in_braces(arg, temp, 0);
+    substitute_vars(temp, temp, ses);
+    fh=open_logfile(ses, temp,
+        "#DUMPING CONSOLE TO {%s}",
+        "#APPENDING CONSOLE DUMP TO {%s}",
+        "#PIPING CONSOLE DUMP TO {%s}");
+    if (fh)
     {
-        tintin_eprintf(ses, "#UI: No console buffer => can't dump it.");
-        return;
+        user_condump(fh);
+        fclose(fh);
     }
-    if (*arg)
-    {
-        arg = get_arg_in_braces(arg, temp, 0);
-        substitute_vars(temp, fname);
-        substitute_myvars(fname, temp, ses);
-        fh=open_logfile(ses, fname,
-            "#DUMPING CONSOLE TO {%s}",
-            "#APPENDING CONSOLE DUMP TO {%s}",
-            "#PIPING CONSOLE DUMP TO {%s}");
-        if (fh)
-        {
-            user_condump(fh);
-            fclose(fh);
-        }
-    }
-    else
-        tintin_eprintf(ses, "#Syntax: #condump <file>");
 }
 
 /********************/
@@ -438,43 +430,39 @@ void condump_command(const char *arg, struct session *ses)
 /********************/
 void log_command(const char *arg, struct session *ses)
 {
-    char fname[BUFFER_SIZE], temp[BUFFER_SIZE];
+    char temp[BUFFER_SIZE];
 
-    if (ses!=nullsession)
+    if (ses==nullsession)
+        return tintin_eprintf(ses, "#THERE'S NO SESSION TO LOG.");
+
+    if (*arg)
     {
-        if (*arg)
-        {
-            if (ses->logfile)
-            {
-                log_off(ses);
-                if (ses->mesvar[MSG_LOG])
-                    tintin_printf(ses, "#OK. LOGGING TURNED OFF.");
-            }
-            get_arg_in_braces(arg, temp, 1);
-            substitute_vars(temp, fname);
-            substitute_myvars(fname, temp, ses);
-            ses->logfile=open_logfile(ses, temp,
-                "#OK. LOGGING TO {%s} .....",
-                "#OK. APPENDING LOG TO {%s} .....",
-                "#OK. PIPING LOG TO {%s} .....");
-            if (ses->logfile)
-                ses->logname=mystrdup(temp);
-            if (!new_conv(&ses->c_log, logcs_charset(ses->logcharset), 1))
-                tintin_eprintf(ses, "#Warning: can't open charset: %s",
-                                    logcs_charset(ses->logcharset));
-
-        }
-        else if (ses->logfile)
+        if (ses->logfile)
         {
             log_off(ses);
             if (ses->mesvar[MSG_LOG])
                 tintin_printf(ses, "#OK. LOGGING TURNED OFF.");
         }
-        else if (ses->mesvar[MSG_LOG])
-            tintin_printf(ses, "#LOGGING ALREADY OFF.");
+        get_arg_in_braces(arg, temp, 1);
+        substitute_vars(temp, temp, ses);
+        ses->logfile=open_logfile(ses, temp,
+            "#OK. LOGGING TO {%s} .....",
+            "#OK. APPENDING LOG TO {%s} .....",
+            "#OK. PIPING LOG TO {%s} .....");
+        if (ses->logfile)
+            ses->logname=mystrdup(temp);
+        if (!new_conv(&ses->c_log, logcs_charset(ses->logcharset), 1))
+            tintin_eprintf(ses, "#Warning: can't open charset: %s",
+                                logcs_charset(ses->logcharset));
     }
-    else
-        tintin_eprintf(ses, "#THERE'S NO SESSION TO LOG.");
+    else if (ses->logfile)
+    {
+        log_off(ses);
+        if (ses->mesvar[MSG_LOG])
+            tintin_printf(ses, "#OK. LOGGING TURNED OFF.");
+    }
+    else if (ses->mesvar[MSG_LOG])
+        tintin_printf(ses, "#LOGGING ALREADY OFF.");
 }
 
 /*************************/
@@ -482,7 +470,7 @@ void log_command(const char *arg, struct session *ses)
 /*************************/
 void debuglog_command(const char *arg, struct session *ses)
 {
-    char fname[BUFFER_SIZE], temp[BUFFER_SIZE];
+    char temp[BUFFER_SIZE];
 
     if (*arg)
     {
@@ -496,8 +484,7 @@ void debuglog_command(const char *arg, struct session *ses)
             ses->debuglogname = NULL;
         }
         get_arg_in_braces(arg, temp, 1);
-        substitute_vars(temp, fname);
-        substitute_myvars(fname, temp, ses);
+        substitute_vars(temp, temp, ses);
         ses->debuglogfile=open_logfile(ses, temp,
             "#OK. DEBUGLOG SET TO {%s} .....",
             "#OK. DEBUGLOG APPENDING TO {%s} .....",
@@ -522,31 +509,29 @@ void debuglog(struct session *ses, const char *format, ...)
 {
     va_list ap;
     char buf[BUFFER_SIZE];
-    struct timeval tv;
 
     if (!ses->debuglogfile)
         return;
 
-    gettimeofday(&tv, 0);
+    timens_t t = current_time()-ses->sessionstart;
     va_start(ap, format);
     if (vsnprintf(buf, BUFFER_SIZE-1, format, ap)>BUFFER_SIZE-2)
         buf[BUFFER_SIZE-3]='>';
     va_end(ap);
-    cfprintf(ses->debuglogfile, "%4ld.%06d: %s\n",
-        (long int)tv.tv_sec-ses->sessionstart, (int)tv.tv_usec, buf);
+    cfprintf(ses->debuglogfile, "%4lld.%06lld: %s\n", t/NANO, t%NANO, buf);
 }
 
 
 struct session* do_read(FILE *myfile, const char *filename, struct session *ses)
 {
     char line[BUFFER_SIZE], buffer[BUFFER_SIZE], lstr[BUFFER_SIZE], *cptr, *eptr;
-    bool flag, ignore_lines;
+    bool want_tt_char, ignore_lines;
     int nl;
     mbstate_t cs;
 
     memset(&cs, 0, sizeof(cs));
 
-    flag = !in_read;
+    want_tt_char = !in_read && !tintin_char_set;
     if (!ses->verbose)
         puts_echoing = false;
     if (!in_read)
@@ -572,14 +557,14 @@ struct session* do_read(FILE *myfile, const char *filename, struct session *ses)
         if (!nl++)
             if (line[0]=='#' && line[1]=='!') /* Unix hashbang script */
                 continue;
-        if (flag)
+        if (want_tt_char)
         {
             puts_echoing = ses->verbose || !real_quiet;
             if (is7punct(*line))
                 char_command(line, ses);
             if (!ses->verbose)
                 puts_echoing = false;
-            flag = false;
+            want_tt_char = false;
         }
         for (cptr = line; *cptr && *cptr != '\n' && *cptr!='\r'; cptr++) ;
         *cptr = '\0';
@@ -656,8 +641,7 @@ struct session* read_command(const char *filename, struct session *ses)
     char buffer[BUFFER_SIZE], fname[BUFFER_SIZE], lfname[BUFFER_SIZE];
 
     get_arg_in_braces(filename, buffer, 1);
-    substitute_vars(buffer, fname);
-    substitute_myvars(fname, buffer, ses);
+    substitute_vars(buffer, buffer, ses);
     expand_filename(buffer, fname, lfname);
     if (!*filename)
     {
@@ -674,11 +658,10 @@ struct session* read_command(const char *filename, struct session *ses)
 }
 
 
-#define WFLAG(name, var, org)   if (var!=(org))                                 \
-                                {                                               \
-                                    sprintf(num, "%d", var);                    \
-                                    prepare_for_write(name, num, 0, 0, buffer); \
-                                    cfputs(buffer, myfile);                     \
+#define WFLAG(name, var, org)   if (var!=(org))                         \
+                                {                                       \
+                                    sprintf(num, "%d", var);            \
+                                    cfcom(myfile, name, num, 0, 0);     \
                                 }
 #define SFLAG(name, var, org)   WFLAG(name, ses->var, org)
 /**********************/
@@ -690,21 +673,16 @@ void write_command(const char *filename, struct session *ses)
     char buffer[BUFFER_SIZE*4], num[32], fname[BUFFER_SIZE], lfname[BUFFER_SIZE];
     struct listnode *nodeptr, *templist;
     struct routenode *rptr;
+    kbtree_t(str) *sl;
+    kbitr_t itr;
 
     get_arg_in_braces(filename, buffer, 1);
-    substitute_vars(buffer, fname);
-    substitute_myvars(fname, buffer, ses);
+    substitute_vars(buffer, buffer, ses);
     expand_filename(buffer, fname, lfname);
-    if (*filename == '\0')
-    {
-        tintin_eprintf(ses, "#ERROR: syntax is: #write <filename>");
-        return;
-    }
+    if (!*filename)
+        return tintin_eprintf(ses, "#ERROR: syntax is: #write <filename>");
     if (!(myfile = fopen(lfname, "w")))
-    {
-        tintin_eprintf(ses, "#ERROR - COULDN'T OPEN FILE {%s}.", fname);
-        return;
-    }
+        return tintin_eprintf(ses, "#ERROR - COULDN'T OPEN FILE {%s}.", fname);
 
     WFLAG("keypad", keypad, DEFAULT_KEYPAD);
     WFLAG("retain", retain, DEFAULT_RETAIN);
@@ -729,67 +707,57 @@ void write_command(const char *filename, struct session *ses)
     SFLAG("messages paths", mesvar[MSG_PATH], DEFAULT_PATH_MESS);
     SFLAG("messages errors", mesvar[MSG_ERROR], DEFAULT_ERROR_MESS);
     SFLAG("messages hooks", mesvar[MSG_HOOK], DEFAULT_HOOK_MESS);
+    SFLAG("messages log", mesvar[MSG_LOG], DEFAULT_LOG_MESS);
+    SFLAG("messages ticks", mesvar[MSG_TICK], DEFAULT_TICK_MESS);
     SFLAG("verbatim", verbatim, false);
-    SFLAG("ticksize", tick_size, DEFAULT_TICK_SIZE);
-    SFLAG("pretick", pretick, DEFAULT_PRETICK);
+    if (ses->tick_size != DEFAULT_TICK_SIZE*NANO)
+    {
+        usecstr(num, ses->tick_size);
+        cfprintf(myfile, "%cticksize %s\n", tintin_char, num);
+    }
+    if (ses->pretick != DEFAULT_PRETICK*NANO)
+    {
+        usecstr(num, ses->pretick);
+        cfprintf(myfile, "%cpretick %s\n", tintin_char, num);
+    }
     if (strcmp(DEFAULT_CHARSET, ses->charset))
         cfprintf(myfile, "%ccharset {%s}\n", tintin_char, ses->charset);
     if (strcmp(logcs_name(DEFAULT_LOGCHARSET), logcs_name(ses->logcharset)))
         cfprintf(myfile, "%clogcharset {%s}\n", tintin_char, logcs_name(ses->logcharset));
     nodeptr = templist = hash2list(ses->aliases, "*");
     while ((nodeptr = nodeptr->next))
-    {
-        prepare_for_write("alias", nodeptr->left, nodeptr->right, 0, buffer);
-        cfputs(buffer, myfile);
-    }
+        cfcom(myfile, "alias", nodeptr->left, nodeptr->right, 0);
     zap_list(templist);
 
     nodeptr = ses->actions;
     while ((nodeptr = nodeptr->next))
-    {
-        prepare_for_write("action", nodeptr->left, nodeptr->right, nodeptr->pr,
-                          buffer);
-        cfputs(buffer, myfile);
-    }
+        cfcom(myfile, "action", nodeptr->left, nodeptr->right, nodeptr->pr);
 
-    nodeptr = ses->antisubs;
-    while ((nodeptr = nodeptr->next))
-    {
-        prepare_for_write("antisub", nodeptr->left, 0, 0, buffer);
-        cfputs(buffer, myfile);
-    }
+    sl = ses->antisubs;
+    for (kb_itr_first(str, sl, &itr); kb_itr_valid(&itr); kb_itr_next(str, sl, &itr))
+        cfcom(myfile, "antisub", kb_itr_key(char*, &itr), 0, 0);
 
     nodeptr = ses->subs;
     while ((nodeptr = nodeptr->next))
     {
         if (strcmp( nodeptr->right, EMPTY_LINE))
-            prepare_for_write("sub", nodeptr->left, nodeptr->right, 0, buffer);
+            cfcom(myfile, "sub", nodeptr->left, nodeptr->right, 0);
         else
-            prepare_for_write("gag", nodeptr->left, 0, 0, buffer);
-        cfputs(buffer, myfile);
+            cfcom(myfile, "gag", nodeptr->left, 0, 0);
     }
 
     nodeptr = templist = hash2list(ses->myvars, "*");
     while ((nodeptr = nodeptr->next))
-    {
-        prepare_for_write("var", nodeptr->left, nodeptr->right, "\0", buffer);
-        cfputs(buffer, myfile);
-    }
+        cfcom(myfile, "var", nodeptr->left, nodeptr->right, 0);
     zap_list(templist);
 
     nodeptr = ses->highs;
     while ((nodeptr = nodeptr->next))
-    {
-        prepare_for_write("highlight", nodeptr->right, nodeptr->left, "\0", buffer);
-        cfputs(buffer, myfile);
-    }
+        cfcom(myfile, "highlight", nodeptr->right, nodeptr->left, 0);
 
     nodeptr = templist = hash2list(ses->pathdirs, "*");
     while ((nodeptr = nodeptr->next))
-    {
-        prepare_for_write("pathdir", nodeptr->left, nodeptr->right, 0, buffer);
-        cfputs(buffer, myfile);
-    }
+        cfcom(myfile, "pathdir", nodeptr->left, nodeptr->right, 0);
     zap_list(templist);
 
     for (int nr=0;nr<MAX_LOCATIONS;nr++)
@@ -809,18 +777,12 @@ void write_command(const char *filename, struct session *ses)
 
     nodeptr = templist = hash2list(ses->binds, "*");
     while ((nodeptr = nodeptr->next))
-    {
-        prepare_for_write("bind", nodeptr->left, nodeptr->right, 0, buffer);
-        cfputs(buffer, myfile);
-    }
+        cfcom(myfile, "bind", nodeptr->left, nodeptr->right, 0);
     zap_list(templist);
 
     for (int nr=0;nr<NHOOKS;nr++)
         if (ses->hooks[nr])
-        {
-            prepare_for_write("hook", hook_names[nr], ses->hooks[nr], 0, buffer);
-            cfputs(buffer, myfile);
-        }
+            cfcom(myfile, "hook", hook_names[nr], ses->hooks[nr], 0);
 
     fclose(myfile);
     tintin_printf(ses, "#COMMANDS-FILE WRITTEN.");
@@ -860,27 +822,19 @@ void writesession_command(const char *filename, struct session *ses)
     char buffer[BUFFER_SIZE*4], *val, num[32], fname[BUFFER_SIZE], lfname[BUFFER_SIZE];
     struct listnode *nodeptr, *onptr;
     struct routenode *rptr;
+    kbtree_t(str) *sl, *orgsl;
+    kbitr_t itr;
 
     if (ses==nullsession)
-    {
-        write_command(filename, ses);
-        return;
-    }
+        return tintin_eprintf(ses, "#THIS IS THE NULL SESSION -- NO DELTA!");
 
     get_arg_in_braces(filename, buffer, 1);
-    substitute_vars(buffer, fname);
-    substitute_myvars(fname, buffer, ses);
+    substitute_vars(buffer, buffer, ses);
     expand_filename(buffer, fname, lfname);
-    if (*filename == '\0')
-    {
-        tintin_eprintf(ses, "#ERROR - COULDN'T OPEN FILE {%s}.", fname);
-        return;
-    }
+    if (!*filename)
+        return tintin_eprintf(ses, "#ERROR - COULDN'T OPEN FILE {%s}.", fname);
     if (!(myfile = fopen(lfname, "w")))
-    {
-        tintin_eprintf(ses, "#ERROR - COULDN'T OPEN FILE {%s}.", fname);
-        return;
-    }
+        return tintin_eprintf(ses, "#ERROR - COULDN'T OPEN FILE {%s}.", fname);
 
 #undef SFLAG
 #define SFLAG(name, var, dummy) WFLAG(name, ses->var, nullsession->var);
@@ -904,9 +858,19 @@ void writesession_command(const char *filename, struct session *ses)
     SFLAG("messages paths", mesvar[MSG_PATH], DEFAULT_PATH_MESS);
     SFLAG("messages errors", mesvar[MSG_ERROR], DEFAULT_ERROR_MESS);
     SFLAG("messages hooks", mesvar[MSG_HOOK], DEFAULT_HOOK_MESS);
+    SFLAG("messages log", mesvar[MSG_LOG], DEFAULT_LOG_MESS);
+    SFLAG("messages ticks", mesvar[MSG_TICK], DEFAULT_TICK_MESS);
     SFLAG("verbatim", verbatim, false);
-    SFLAG("ticksize", tick_size, DEFAULT_TICK_SIZE);
-    SFLAG("pretick", pretick, DEFAULT_PRETICK);
+    if (ses->tick_size != DEFAULT_TICK_SIZE*NANO)
+    {
+        usecstr(num, ses->tick_size);
+        cfprintf(myfile, "%cticksize %s\n", tintin_char, num);
+    }
+    if (ses->pretick != DEFAULT_PRETICK*NANO)
+    {
+        usecstr(num, ses->pretick);
+        cfprintf(myfile, "%cpretick %s\n", tintin_char, num);
+    }
     if (strcmp(nullsession->charset, ses->charset))
         cfprintf(myfile, "%ccharset {%s}\n", tintin_char, ses->charset);
     if (strcmp(logcs_name(nullsession->logcharset), logcs_name(ses->logcharset)))
@@ -918,8 +882,7 @@ void writesession_command(const char *filename, struct session *ses)
         if ((val=get_hash(nullsession->aliases, nodeptr->left)))
             if (!strcmp(val, nodeptr->right))
                 continue;
-        prepare_for_write("alias", nodeptr->left, nodeptr->right, 0, buffer);
-        cfputs(buffer, myfile);
+        cfcom(myfile, "alias", nodeptr->left, nodeptr->right, 0);
     }
     zap_list(onptr);
 
@@ -930,19 +893,16 @@ void writesession_command(const char *filename, struct session *ses)
             if (!strcmp(onptr->right, nodeptr->right)||
                     !strcmp(onptr->right, nodeptr->right))
                 continue;
-        prepare_for_write("action", nodeptr->left, nodeptr->right, nodeptr->pr,
-                          buffer);
-        cfputs(buffer, myfile);
+        cfcom(myfile, "action", nodeptr->left, nodeptr->right, nodeptr->pr);
     }
 
-    nodeptr = ses->antisubs;
-    while ((nodeptr = nodeptr->next))
+    sl = ses->antisubs;
+    orgsl = nullsession->antisubs;
+    for (kb_itr_first(str, sl, &itr); kb_itr_valid(&itr); kb_itr_next(str, sl, &itr))
     {
-        if ((onptr=searchnode_list(nullsession->antisubs, nodeptr->left)))
-            if (!strcmp(onptr->right, nodeptr->right))
-                continue;
-        prepare_for_write("antisubstitute", nodeptr->left, 0, 0, buffer);
-        cfputs(buffer, myfile);
+        char *p = kb_itr_key(char*, &itr);
+        if (!kb_get(str, orgsl, p))
+            cfcom(myfile, "antisub", p, 0, 0);
     }
 
     nodeptr = ses->subs;
@@ -952,10 +912,9 @@ void writesession_command(const char *filename, struct session *ses)
             if (!strcmp(onptr->right, nodeptr->right))
                 continue;
         if (strcmp( nodeptr->right, EMPTY_LINE))
-            prepare_for_write("sub", nodeptr->left, nodeptr->right, 0, buffer);
+            cfcom(myfile, "sub", nodeptr->left, nodeptr->right, 0);
         else
-            prepare_for_write("gag", nodeptr->left, 0, 0, buffer);
-        cfputs(buffer, myfile);
+            cfcom(myfile, "gag", nodeptr->left, 0, 0);
     }
 
     nodeptr = onptr = hash2list(ses->myvars, "*");
@@ -964,8 +923,7 @@ void writesession_command(const char *filename, struct session *ses)
         if ((val=get_hash(nullsession->myvars, nodeptr->left)))
             if (!strcmp(val, nodeptr->right))
                 continue;
-        prepare_for_write("var", nodeptr->left, nodeptr->right, 0, buffer);
-        cfputs(buffer, myfile);
+        cfcom(myfile, "var", nodeptr->left, nodeptr->right, 0);
     }
     zap_list(onptr);
 
@@ -975,8 +933,7 @@ void writesession_command(const char *filename, struct session *ses)
         if ((onptr=searchnode_list(nullsession->highs, nodeptr->left)))
             if (!strcmp(onptr->right, nodeptr->right))
                 continue;
-        prepare_for_write("highlight", nodeptr->right, nodeptr->left, 0, buffer);
-        cfputs(buffer, myfile);
+        cfcom(myfile, "highlight", nodeptr->right, nodeptr->left, 0);
     }
 
     nodeptr = onptr = hash2list(ses->pathdirs, "*");
@@ -985,8 +942,7 @@ void writesession_command(const char *filename, struct session *ses)
         if ((val=get_hash(nullsession->pathdirs, nodeptr->left)))
             if (!strcmp(val, nodeptr->right))
                 continue;
-        prepare_for_write("pathdirs", nodeptr->left, nodeptr->right, 0, buffer);
-        cfputs(buffer, myfile);
+        cfcom(myfile, "pathdirs", nodeptr->left, nodeptr->right, 0);
     }
     zap_list(onptr);
 
@@ -1017,8 +973,7 @@ void writesession_command(const char *filename, struct session *ses)
         if ((val=get_hash(nullsession->binds, nodeptr->left)))
             if (!strcmp(val, nodeptr->right))
                 continue;
-        prepare_for_write("bind", nodeptr->left, nodeptr->right, 0, buffer);
-        cfputs(buffer, myfile);
+        cfcom(myfile, "bind", nodeptr->left, nodeptr->right, 0);
     }
     zap_list(onptr);
 
@@ -1026,27 +981,28 @@ void writesession_command(const char *filename, struct session *ses)
         if (ses->hooks[nr])
             if (!nullsession->hooks[nr] ||
                 strcmp(ses->hooks[nr], nullsession->hooks[nr]))
-                {
-                    prepare_for_write("hook", hook_names[nr], ses->hooks[nr], 0, buffer);
-                    cfputs(buffer, myfile);
-                }
+            {
+                cfcom(myfile, "hook", hook_names[nr], ses->hooks[nr], 0);
+            }
 
     fclose(myfile);
-    tintin_printf(ses, "#COMMANDS-FILE WRITTEN.");
+    tintin_printf(ses, "#DELTA COMMANDS FILE WRITTEN.");
     return;
 }
 
 
-static void prepare_for_write(const char *command, const char *left, const char *right, const char *pr, char *result)
+static void cfcom(FILE *f, const char *command, const char *left, const char *right, const char *pr)
 {
+    char buffer[BUFFER_SIZE*4], *b = buffer;
     /* "result" is four times as big as the regular buffer.  This is */
     /* pointless as read line are capped at 1/2 buffer anyway.       */
-    result+=sprintf(result, "%c%s {%s}", tintin_char, command, left);
+    b+=sprintf(b, "%c%s {%s}", tintin_char, command, left);
     if (right)
-        result+=sprintf(result, " {%s}", right);
+        b+=sprintf(b, " {%s}", right);
     if (pr && strlen(pr))
-        result+=sprintf(result, " {%s}", pr);
-    sprintf(result, "\n");
+        b+=sprintf(b, " {%s}", pr);
+    *b++='\n'; *b=0;
+    cfputs(buffer, f);
 }
 
 
@@ -1062,19 +1018,15 @@ void textin_command(const char *arg, struct session *ses)
     memset(&cs, 0, sizeof(cs));
 
     get_arg_in_braces(arg, buffer, 1);
-    substitute_vars(buffer, filename);
-    substitute_myvars(filename, buffer, ses);
+    substitute_vars(buffer, buffer, ses);
     expand_filename(buffer, filename, lfname);
     if (ses == nullsession)
     {
-        tintin_eprintf(ses, "#You can't read any text in without a session being active.");
-        return;
+        return tintin_eprintf(ses,
+            "#You can't read any text in without a session being active.");
     }
     if (!(myfile = fopen(lfname, "r")))
-    {
-        tintin_eprintf(ses, "ERROR: File {%s} doesn't exist.", filename);
-        return;
-    }
+        return tintin_eprintf(ses, "ERROR: File {%s} doesn't exist.", filename);
     while (fgets(buffer, sizeof(buffer), myfile))
     {
         for (cptr = buffer; *cptr && *cptr != '\n'; cptr++) ;
@@ -1102,10 +1054,7 @@ void logtype_command(const char *arg, struct session *ses)
 
     arg=get_arg(arg, left, 1, ses);
     if (!*left)
-    {
-        tintin_printf(ses, "#The log type is: %s", logtypes[ses->logtype]);
-        return;
-    }
+        return tintin_printf(ses, "#The log type is: %s", logtypes[ses->logtype]);
     for (unsigned t=0;t<sizeof(logtypes)/sizeof(char*);t++)
         if (is_abrev(left, logtypes[t]))
         {
@@ -1129,19 +1078,13 @@ void logcharset_command(const char *arg, struct session *ses)
     cset=what;
 
     if (!*cset)
-    {
-        tintin_printf(ses, "#Log charset: %s", logcs_name(ses->logcharset));
-        return;
-    }
+        return tintin_printf(ses, "#Log charset: %s", logcs_name(ses->logcharset));
     if (!strcasecmp(cset, "local"))
         cset=LOGCS_LOCAL;
     else if (!strcasecmp(cset, "remote"))
         cset=LOGCS_REMOTE;
     if (!new_conv(&nc, logcs_charset(cset), 1))
-    {
-        tintin_eprintf(ses, "#No such charset: {%s}", logcs_charset(cset));
-        return;
-    }
+        return tintin_eprintf(ses, "#No such charset: {%s}", logcs_charset(cset));
     if (!logcs_is_special(ses->logcharset))
         SFREE(ses->logcharset);
     ses->logcharset=logcs_is_special(cset) ? cset : mystrdup(cset);
