@@ -12,12 +12,12 @@
 #include "protos/hash.h"
 #include "protos/hooks.h"
 #include "protos/print.h"
-#include "protos/prof.h"
 #include "protos/net.h"
 #include "protos/parse.h"
 #include "protos/path.h"
 #include "protos/utils.h"
-#include "protos/variables.h"
+#include "protos/string.h"
+#include "protos/vars.h"
 
 #include "commands.h"
 typedef void (*t_command)(const char*, struct session*);
@@ -36,39 +36,39 @@ static inline bool is_speedwalk_dirs(const char *cp);
 
 static inline const char *get_arg_stop_spaces(const char *s, char *arg);
 static const char *get_arg_all(const char *s, char *arg);
-bool in_alias=false;
 static struct hashtable *commands, *c_commands;
-int recursion;
+
+bool inc_recursion(void)
+{
+    if (++recursion>=MAX_RECURSION)
+    {
+        if (recursion==MAX_RECURSION)
+            tintin_eprintf(0, "#TOO DEEP RECURSION.");
+        recursion=MAX_RECURSION*3;
+        return true;
+    }
+
+    return false;
+}
 
 /**************************************************************************/
 /* parse input, check for TINTIN commands and aliases and send to session */
 /**************************************************************************/
 struct session* parse_input(const char *input, bool override_verbatim, struct session *ses)
 {
-    char command[BUFFER_SIZE], arg[BUFFER_SIZE], result[BUFFER_SIZE], *al;
+    char command[BUFFER_SIZE], arg[BUFFER_SIZE], *al;
     int nspaces;
-#ifdef PROFILING
-    const char *oldprof=prof_area;
-    PROF("parsing input");
-# define PPOP prof_area=oldprof
-#else
-# define PPOP
-#endif
 
-    if (++recursion>=MAX_RECURSION)
+    if (inc_recursion())
     {
         in_alias=false;
-        if (recursion==MAX_RECURSION)
-            tintin_eprintf(ses, "#TOO DEEP RECURSION.");
-        recursion=MAX_RECURSION*3;
-        PPOP;
         return ses;
     }
 
     debuglog(ses, "%s", input);
     if (!ses->server_echo && activesession == ses)
         term_echoing = true;
-    if (*input == '\0')
+    if (!*input)
     {
         if (ses!=nullsession)
         {
@@ -77,66 +77,42 @@ struct session* parse_input(const char *input, bool override_verbatim, struct se
         }
         else if (!in_read)
             write_com_arg_mud("", "", 0, ses);
-        PPOP;
         return ses;
     }
     if ((*input==tintin_char) && is_abrev(input + 1, "verbatim"))
     {
         verbatim_command("", ses);
         debuglog(ses, "%s", input);
-        PPOP;
         return ses;
     }
     if (ses->verbatim && !override_verbatim && (ses!=nullsession))
     {
         write_line_mud(input, ses);
         debuglog(ses, "%s", input);
-        PPOP;
         return ses;
     }
-    if (*input==verbatim_char && (ses!=nullsession))
+    if (*input==VERBATIM_CHAR && (ses!=nullsession))
     {
         input++;
         write_line_mud(input, ses);
         debuglog(ses, "%s", input-1);
-        PPOP;
         return ses;
     }
 
     while (*input)
     {
-        PROFPUSH("expanding variables");
         while (*input == ';')
             input=space_out(input+1);
-        if (pvars)
+        input = get_command(input, command);
+        substitute_vars(command, command, ses);
+        nspaces=0;
+        while (*input==' ')
         {
-            input = get_command(input, command);
-            substitute_vars( command, result);
-            substitute_myvars( result, command, ses);
-            nspaces=0;
-            while (*input==' ')
-            {
-                input++;
-                nspaces++;
-            }
-            input = get_arg_all(input, arg);
-            substitute_vars( arg, result);
-            substitute_myvars( result, arg, ses);
+            input++;
+            nspaces++;
         }
-        else
-        {
-            input = get_command(input, result);
-            substitute_myvars( result, command, ses);
-            nspaces=0;
-            while (*input==' ')
-            {
-                input++;
-                nspaces++;
-            }
-            input = get_arg_all(input, result);
-            substitute_myvars( result, arg, ses);
-        }
-        PROFPOP;
+        input = get_arg_all(input, arg);
+        substitute_vars(arg, arg, ses);
 
         if (in_alias)
         {
@@ -157,15 +133,11 @@ struct session* parse_input(const char *input, bool override_verbatim, struct se
             in_alias=false;
         }
         if (recursion>MAX_RECURSION)
-        {
-            PPOP;
             return ses;
-        }
         if (aborting)
         {
             aborting=false;
             recursion--;
-            PPOP;
             return ses;
         }
         debuglog(ses, "%s", command);
@@ -179,7 +151,6 @@ struct session* parse_input(const char *input, bool override_verbatim, struct se
 
             strcpy(vars[0], arg);
 
-            PROF("expanding aliases");
             for (int i = 1; i < 10; i++)
                 cpsource=get_arg_in_braces(cpsource, vars[i], 0);
             in_alias=true;
@@ -196,14 +167,12 @@ struct session* parse_input(const char *input, bool override_verbatim, struct se
             if ((*arg)||!do_goto(command, ses))
 #endif
             {
-                PROF("expanding text being sent");
                 get_arg_with_spaces(arg, arg);
                 write_com_arg_mud(command, arg, nspaces, ses);
             }
     }
     aborting=false;
     recursion--;
-    PPOP;
     return ses;
 }
 
@@ -302,10 +271,6 @@ static struct session* parse_tintin_command(const char *command, const char *arg
 {
     const char *func, *a;
     char *b, cmd[BUFFER_SIZE], right[BUFFER_SIZE];
-#ifdef PROFILING
-    const char *oldprof=prof_area;
-    PROF("executing commands");
-#endif
 
     for (struct session *sesptr = sessionlist; sesptr; sesptr = sesptr->next)
         if (strcmp(sesptr->name, command) == 0)
@@ -314,7 +279,6 @@ static struct session* parse_tintin_command(const char *command, const char *arg
             {
                 get_arg_in_braces(arg, right, 1);
                 parse_input(right, true, sesptr);     /* was: #sessioname commands */
-                PPOP;
                 return ses;
             }
             else
@@ -322,7 +286,6 @@ static struct session* parse_tintin_command(const char *command, const char *arg
                 activesession = sesptr;
                 tintin_printf(ses, "#SESSION '%s' ACTIVATED.", sesptr->name);
                 do_hook(ses, HOOK_DEACTIVATE, 0, true);  /* FIXME: blockzap */
-                PPOP;
                 return do_hook(sesptr, HOOK_ACTIVATE, 0, false);
             }
         }
@@ -345,7 +308,6 @@ static struct session* parse_tintin_command(const char *command, const char *arg
         }
         else
             tintin_eprintf(ses, "#Cannot repeat a command a non-positive number of times.");
-        PPOP;
         return ses;
     }
     else if ((func=get_hash(c_commands, cmd)))
@@ -354,7 +316,6 @@ static struct session* parse_tintin_command(const char *command, const char *arg
         ((t_command)func)(arg, ses);
     else
         tintin_eprintf(ses, "#UNKNOWN TINTIN-COMMAND: [%c%s]", tintin_char, command);
-    PPOP;
     return ses;
 }
 
@@ -382,9 +343,18 @@ void init_parse(void)
     c_commands=init_hash();
     set_hash_nostring(commands, "end", (char*)end_command);
     set_hash_nostring(commands, "unlink", (char*)unlink_command);
+/* this terrible cast avoid gcc (-w → -Wcast-function-type) insisting it
+   still knows better despite an explicit cast */
+#define SC t_command)(void(*)(void)
 #include "load_commands.h"
 }
 
+void cleanup_parse(void)
+{
+    kill_hash_nostring(commands);
+    kill_hash_nostring(c_commands);
+}
+
 
 /**********************************************/
 /* get all arguments - don't remove "s and \s */
@@ -515,7 +485,7 @@ const char* get_arg_in_braces(const char *s, char *arg, bool allow_spaces)
         return s;
     }
     s++;
-    while (*s != '\0' && !(*s == BRACE_CLOSE && nest == 0))
+    while (*s && !(*s == BRACE_CLOSE && nest == 0))
     {
         if (*s==CHAR_VERBATIM)     /* \ */
             ;
@@ -569,12 +539,7 @@ static inline const char* get_arg_stop_spaces(const char *s, char *arg)
 const char* get_arg(const char *s, char *arg, bool allow_spaces, struct session *ses)
 {
     const char *cptr=get_arg_in_braces(s, arg, allow_spaces);
-    if (*arg)
-    {
-        char tmp[BUFFER_SIZE];
-        substitute_vars(arg, tmp);
-        substitute_myvars(tmp, arg, ses);
-    }
+    substitute_vars(arg, arg, ses);
     return cptr;
 }
 
