@@ -7,11 +7,14 @@
 #include "protos/unicode.h"
 #include "protos/user.h"
 #include "protos/utils.h"
-#if HAVE_TERMIOS_H
+#ifdef HAVE_TERMIOS_H
 # include <termios.h>
 #endif
 #include <assert.h>
 #include <sys/ioctl.h>
+#ifdef HAVE_VALGRIND_VALGRIND_H
+#include <valgrind/valgrind.h>
+#endif
 
 static mbstate_t outstate;
 #define OUTSTATE &outstate
@@ -24,6 +27,7 @@ static int o_len, o_pos, o_oldcolor, o_prevcolor, o_draftlen, o_lastprevcolor;
 #define o_color color
 #define o_lastcolor lastcolor
 static int b_first, b_current, b_last, b_bottom, b_screenb;
+static int b_pending_newline;
 static bool o_strongdraft;
 static int b_greeting;
 static char *b_output[CONSOLE_LENGTH];
@@ -283,10 +287,19 @@ static void redraw_status(void)
         if (getcolor(&pos, &color, 0))
         {
             c=color;
-            if (!(c&0x70))
-                c=c|(STATUS_COLOR<<4);
-            if ((c&15)==((c>>4)&7))
-                c=(c&0xf0)|(c&7? 0:(STATUS_COLOR==COLOR_BLACK? 7:0));
+            if (!(c&CBG_MASK))
+                c=c|(STATUS_COLOR<<CBG_AT);
+            if ((c&CFG_MASK) == (c&CBG_MASK) >> CBG_AT)
+            {
+                int k = c&CFG_MASK;
+                if (k != 0 && k != 8)
+                    k = 0;
+                else if (STATUS_COLOR == COLOR_BLACK)
+                    k = 7;
+                else
+                    k = 0;
+                c=(c&~CFG_MASK)|k;
+            }
             tbuf=ansicolor(tbuf, c);
             pos++;
         }
@@ -370,8 +383,18 @@ static void b_scroll(int b_to)
     term_commit();
 }
 
+static void b_flush_newline(void)
+{
+    if (!b_pending_newline)
+        return;
+    tbuf+=sprintf(tbuf, "\033[0;37;40m\r\n\033[2K");
+    b_pending_newline=0;
+}
+
 static void b_addline(void)
 {
+    b_flush_newline();
+    b_pending_newline=1;
     char *new;
     while (!(new=MALLOC(o_len+1)))
         if (!b_shorten())
@@ -380,7 +403,7 @@ static void b_addline(void)
     strcpy(new, out_line);
     if (b_bottom==b_first+CONSOLE_LENGTH)
         b_shorten();
-    b_output[(unsigned int)b_current%(unsigned int)CONSOLE_LENGTH]=new;
+    b_output[b_current%CONSOLE_LENGTH]=new;
     o_pos=0;
     o_len=setcolor(out_line, o_oldcolor=o_color);
     if (b_bottom<++b_current)
@@ -389,7 +412,6 @@ static void b_addline(void)
         if (b_current==b_screenb+1)
         {
             ++b_screenb;
-            draw_out(out_line);
             term_commit();
         }
         else
@@ -420,13 +442,14 @@ static inline void print_char(const WC ch)
     if (o_pos+dw-1>=COLS)
     {
         out_line[o_len++]='\r';
-        tbuf+=sprintf(tbuf, "\033[0;37;40m\r\n\033[2K");
         b_addline();
+        b_flush_newline();
+        tbuf=ansicolor(tbuf, o_color);
     }
     else if (o_oldcolor!=o_color) /* b_addline already updates the color */
     {
-        if ((o_color&15)==((o_color&0x70)>>4))
-            o_color=(o_color&0xf0)|((o_color&7)?0:7);
+        if ((o_color&CFG_MASK)==((o_color&CBG_MASK)>>CBG_AT))
+            o_color=(o_color&~CFG_MASK)|((o_color&CFG_MASK)?0:7);
         o_len+=setcolor(out_line+o_len, o_color);
         if (b_screenb==b_bottom)
             tbuf=ansicolor(tbuf, o_color);
@@ -444,10 +467,7 @@ static inline void print_char(const WC ch)
 static void form_feed(void)
 {
     for (int i=(isstatus?2:1);i<LINES;i++)
-    {
-        tbuf+=sprintf(tbuf, "\033[0;37;40m\r\n\033[2K");
         b_addline();
-    }
     tbuf+=sprintf(tbuf, "\033[f");
 }
 
@@ -457,6 +477,7 @@ static void b_textout(const char *txt)
 
     /* warning! terminal output can get discarded! */
     *tbuf++=27, *tbuf++='8';
+    b_flush_newline();
     tbuf=ansicolor(tbuf, o_color);
     for (;*txt;txt++)
         switch (*txt)
@@ -474,7 +495,6 @@ static void b_textout(const char *txt)
             break;
         case '\n':
             out_line[o_len]=0;
-            tbuf+=sprintf(tbuf, "\033[0;37;40m\r\n\033[2K");
             b_addline();
             break;
         case 9:
@@ -922,6 +942,8 @@ static bool usertty_process_kbd(struct session *ses, WC ch)
             bind_xterm((val[0]==0 && val[1]==115) /* konsole */
                        || val[0]==1               /* libvte, mlterm, aterm, termit */
                        || val[0]==41              /* xterm, terminology */
+                       || val[0]==64              /* zutty */
+                       || val[0]==83              /* screen */
                       );
         }
         else
@@ -1580,9 +1602,9 @@ static void usertty_init(void)
     in_getpassword=false;
 
     b_first=0;
-    b_current=-1;
-    b_bottom=-1;
-    b_screenb=-1;
+    b_current=0;
+    b_bottom=0;
+    b_screenb=0;
     b_last=-1;
     b_output[0]=out_line;
     out_line[0]=0;
@@ -1612,16 +1634,25 @@ static void usertty_init(void)
 static void usertty_done(void)
 {
     ui_own_output=false;
+    b_flush_newline();
     tbuf+=sprintf(tbuf, "\033[1;%dr\033[%d;1f\033[?25h\033[?7h\033[0;37;40m", LINES, LINES);
     usertty_keypad(false);
     term_commit();
     term_restore();
     write_stdout("\n", 1);
+#ifdef HAVE_VALGRIND_VALGRIND_H
+    if (RUNNING_ON_VALGRIND)
+    {
+        while (b_shorten())
+            ;
+    }
+#endif
 }
 
 static void usertty_pause(void)
 {
     usertty_keypad(false);
+    b_flush_newline();
     tbuf+=sprintf(tbuf, "\033[1;%dr\033[%d;1f\033[?25h\033[?7h\033[0;37;40m", LINES, LINES);
     term_commit();
     term_restore();
